'use strict';

/* jshint quotmark: double */
window.SwaggerTranslator.learn({
    "Warning: Deprecated":"Attenzione: Deprecato",
    "Implementation Notes":"Note di implementazione",
    "Response Class":"Classe della risposta",
    "Status":"Stato",
    "Parameters":"Parametri",
    "Parameter":"Parametro",
    "Value":"Valore",
    "Description":"Descrizione",
    "Parameter Type":"Tipo di parametro",
    "Data Type":"Tipo di dato",
    "Response Messages":"Messaggi della risposta",
    "HTTP Status Code":"Codice stato HTTP",
    "Reason":"Motivo",
    "Response Model":"Modello di risposta",
    "Request URL":"URL della richiesta",
    "Response Body":"Corpo della risposta",
    "Response Code":"Oggetto della risposta",
    "Response Headers":"Intestazioni della risposta",
    "Hide Response":"Nascondi risposta",
    "Try it out!":"Provalo!",
    "Show/Hide":"Mostra/Nascondi",
    "List Operations":"Mostra operazioni",
    "Expand Operations":"Espandi operazioni",
    "Raw":"Grezzo (raw)",
    "can't parse JSON.  Raw result":"non è possibile parsare il JSON. Risultato grezzo (raw).",
    "Model Schema":"Schema del modello",
    "Model":"Modello",
    "apply":"applica",
    "Username":"Nome utente",
    "Password":"Password",
    "Terms of service":"Condizioni del servizio",
    "Created by":"Creato da",
    "See more at":"Informazioni aggiuntive:",
    "Contact the developer":"Contatta lo sviluppatore",
    "api version":"versione api",
    "Response Content Type":"Tipo di contenuto (content type) della risposta",
    "fetching resource":"recuperando la risorsa",
    "fetching resource list":"recuperando lista risorse",
    "Explore":"Esplora",
    "Show Swagger Petstore Example Apis":"Mostra le api di esempio di Swagger Petstore",
    "Can't read from server.  It may not have the appropriate access-control-origin settings.":"Non è possibile leggere dal server. Potrebbe non avere le impostazioni di controllo accesso origine (access-control-origin) appropriate.",
    "Please specify the protocol for":"Si prega di specificare il protocollo per",
    "Can't read swagger JSON from":"Impossibile leggere JSON swagger da:",
    "Finished Loading Resource Information. Rendering Swagger UI":"Lettura informazioni risorse termianta. Swagger UI viene mostrata",
    "Unable to read api":"Impossibile leggere la api",
    "from path":"da cartella",
    "server returned":"il server ha restituito"
});
